/**
  Lighting demo application.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-01-31 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


#include "image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "image.cpp"

#include "glsl.h" /* Orion's tiny GLSL library */
#include "glsl.cpp"

/*********************** App Code **********************/

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Render all geometry in the world */
void draw_the_world(void)
{
	glPushMatrix();
	glTranslatef(+0.5,+0.5,+0.5); glScalef(0.5,0.5,0.5); 
	
	glColor3f(0.5,0.5,0.5); /* gray */
	glutSolidSphere(0.4, 25,50);
	
	glColor3f(1,1,0); /* yellow */
	glutSolidTorus(0.15,0.75, 25,50);
	
	glColor3f(1,0,0); /* red */
	glPushMatrix();
	glRotatef(45.0,1.0,0,0);
	glTranslatef(0,0.7,0.0);
	glutSolidTeapot(0.5);
	glPopMatrix();
	
	/* Per-axis spheres */
	for (int axis=0;axis<3;axis++) {
		glPushMatrix();
		glRotatef(-20.0*axis,1,1,1); /* add a little skew */
		vec3 scale=vec3(0.25); /* squish the sphere */
		scale[axis]=1.0;  /* except this axis */
		glScalef(VEC3_TO_XYZ(scale)); 
		glColor3fv(scale); /* red, green, blue */
		glutSolidSphere(1.0,40,40);
		glPopMatrix();
	}
	glPopMatrix();
}

/* Orientation of our camera. Camera looks down -Z axis; Y is up. */
ortho_frame camera_frame;

int texWid=512,texHt=512;

void my_display(void) {
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	vec3 scene_center=vec3(0.5,0.5,0.5);
	vec3 camera_location=scene_center+0.9*camera_frame.z; /* back away from origin */
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(scene_center),
		VEC3_TO_XYZ(camera_frame.y));
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	/* Shadow texture setup */
	static GLuint shadowTex=0;
	if (!shadowTex) { 
		glGenTextures(1,&shadowTex); /* make texture handle */
		
		glBindTexture(GL_TEXTURE_2D,shadowTex);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
			GL_LINEAR); // _MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		//glTexParameteri(GL_TEXTURE_2D,GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
		
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 
			texWid,texHt, 0, GL_RGBA,GL_UNSIGNED_BYTE,0);
		
		glActiveTexture(GL_TEXTURE3); /* turn on texture "unit" */
		glBindTexture(GL_TEXTURE_2D,shadowTex);
		
		glActiveTexture(GL_TEXTURE0); /* back to default texture "unit" */
	}
	
/* Render the world from POV of the light source 
Output color should equal the geometry depth.
	*/
	static GLhandleARB shadowProg=makeProgramObject(
	"// GLSL Vertex shader\n"
	"varying vec3 shadowCoords;\n"
	"void main(void) {\n" /* FIXME: shadowmap matrix? */
	"	vec4 p = gl_ModelViewMatrix * gl_Vertex; \n"
	"	shadowCoords = vec3( p );\n"
	"	/* map coordinates from [0,1] to [-1,1] along each axis */\n"
	"	gl_Position = p * vec4(2,2,2,1) - vec4(1,1,1,0);\n"
	"}\n"
	,
	"// GLSL Fragment shader\n"
	"varying vec3 shadowCoords;\n"
	"void main(void) {\n"
	"	gl_FragColor = vec4(shadowCoords.z);\n"
	"}\n"
	);
	
	glUseProgramObjectARB(shadowProg);

	glClearColor(1.0,1.0,1.0,0); /* set background color to white (farthest away) */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS); /* pass depth-test if we're closer to the camera (smaller Z) */
	
	draw_the_world();
	
	/* Read the rendered shadow-fiew depths into the shadow texture */
	glBindTexture(GL_TEXTURE_2D,shadowTex);
	glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0,0, 0,0, texWid,texHt);
	
	if (0) { /* Debugging: show the shadow image */
		glutSwapBuffers();
		return;
	}

/* Render world from POV of camera, with ordinary colors. */
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS); /* pass depth-test if we're closer to the camera (smaller Z) */

	static GLhandleARB prog=makeProgramObject(
	"// GLSL Vertex shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"varying vec3 worldNormal;\n"
	"\n"
	"void main(void) {\n"
	"	reflectance = gl_Color;\n"
	"	worldCoords = vec3( gl_ModelViewMatrix * gl_Vertex );\n"
	"	worldNormal = normalize( gl_NormalMatrix * gl_Normal );\n"
	"	gl_Position = gl_ProjectionMatrix * gl_ModelViewMatrix * gl_Vertex; \n"
	"}\n"
	,
	"// GLSL Fragment shader\n"
	"varying vec4 reflectance;\n"
	"varying vec3 worldCoords;\n"
	"varying vec3 worldNormal;\n"
	"uniform vec3 cameraLoc;\n"
	"uniform sampler2D shadowMap;\n"
	"\n"
	"void main(void) {\n"
	"	vec3 wc = step(fract(8.0*worldCoords),vec3(0.5))*2.0-1.0;\n"
	"	float checker = clamp(wc.x*wc.y*wc.z,0.0,1.0);\n"
	"\n"	
	"	vec3 N = normalize(worldNormal); // Points away from surface\n"
	"	vec3 L1 = normalize(vec3(0,0,-1)); // Points toward light 1\n"
	"	vec3 C = normalize(vec3(cameraLoc)-worldCoords); // Points toward camera\n"
	"	vec3 H1 = normalize(C+L1); // Blinn's 'halfway vector' for light 1\n"
	"\n"	
	"	float a=0.2; // Ambient illumination\n"
	"	float d=clamp(dot(N,L1),0.0,1.0); // Diffuse light fraction\n"
	"	float s=pow(clamp(dot(N,H1),0.0,1.0),50.0); // Specular fraction\n"
	"\n"
	"	vec3 shadowCoords = worldCoords;\n"
	"	vec4 shadowMapPix = texture2D(shadowMap,vec2(shadowCoords));\n"
	"	float lit=1.0;\n"
	"	if (shadowCoords.z > shadowMapPix.z + 1.0/256) lit = 0.0; \n"
	"	gl_FragColor = (a+d*lit)*reflectance + vec4(s*lit);\n"
	"	//gl_FragColor = shadowMapPix;\n"
	"}\n"
	);
	glUseProgramObjectARB(prog);
	
	glUniform3fvARB( /* set the "cameraCoords" uniform variable */
		glGetUniformLocationARB(prog, "cameraLoc"),
		1,
		camera_location
	);
	glUniform1iARB(glGetUniformLocationARB(prog,"shadowMap"),3); /* tex unit 3 */
	
	draw_the_world();
	
	glUseProgramObjectARB(0); /* back to fixed-function */
	
	
	
	glutSwapBuffers(); /* copy back buffer to front */

	
/* Statistics */
	static int last_msec=0;
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,-speed*dy); /* OpenGL Y is negative of mouse Y! */
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

int main(int argc,char *argv[]) {
	glutInitWindowSize(texWid,texHt);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	glutCreateWindow("GLSL Lighting Demo");
	glewInit();
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	
	glutMainLoop(); /* processes events */
	return 0;
}
